from copy import deepcopy

import pytest

from ._ctl import SnapCtl
from ._env import SnapEnviron


@pytest.fixture
def snap_env():
    """Environment variables defined in a snap."""
    yield {
        'SNAP': '/snap/mysnap/123',
        'SNAP_COMMON': '/var/snap/mysnap/common',
        'SNAP_DATA': '/var/snap/mysnap/123',
        'SNAP_INSTANCE_NAME': 'mysnap_inst',
        'SNAP_NAME': 'mysnap',
        'SNAP_REVISION': '123',
        'SNAP_USER_COMMON': '/home/user/snap/mysnap/common',
        'SNAP_USER_DATA': '/home/s/snap/mysnap/123',
        'SNAP_VERSION': '0.1.2'
    }


@pytest.fixture
def snap_environ(snap_env):
    """A SnapEnviron using the snap_env."""
    yield SnapEnviron(environ=snap_env)


@pytest.fixture
def snap_apply_env(monkeypatch, snap_env):
    """Apply snap environment variables."""
    for key, value in snap_env.items():
        monkeypatch.setenv(key, value)


@pytest.fixture
def snap_config():
    """A sample snap configuration."""
    yield {
        'foo': 123,
        'bar': 'BAR',
        'baz': {
            'aaa': 'nested',
            'bbb': {
                'ccc': 'more nested'
            }
        },
        'blah': [1, 2, 3]
    }


@pytest.fixture
def snapctl(mocker, snap_apply_env):
    """A SnapCtl instance with a mocked run method."""
    snapctl = SnapCtl(executable='/not/here')
    snapctl.run = mocker.Mock(return_value='')
    yield snapctl


class FakeSnapCtl:
    """A fake SnapCtl implementation."""

    def __init__(self, configs=None, services=None):
        self._configs = configs or {}
        self._services = services or []

    def config_get(self, *keys):
        options = {}
        for key in keys:
            if key in self._configs:
                options[key] = deepcopy(self._configs[key])
        return options

    def config_set(self, configs):
        for key, value in configs.items():
            old_conf = conf = self._configs
            for token in key.split('.'):
                entry = conf.get(token)
                if not isinstance(entry, dict):
                    conf[token] = {}

                old_conf, conf = conf, conf[token]
            old_conf[token] = value

    def services(self):
        return deepcopy(self._services)


@pytest.fixture
def fake_snapctl(snap_config):
    """A fake SnapCtl handling the config."""
    yield FakeSnapCtl(configs=snap_config)



from argparse import (
    ArgumentParser,
    Namespace,
)

from .._hook import get_hooks
from .._snap import Snap
from ._script import Script


class HookScript(Script):
    """A wrapper script to call hooks registered via :data:`pkg_resources`."""

    def get_parser(self) -> ArgumentParser:
        parser = ArgumentParser(
            description='Run the specified hook registered via pkg_resources')
        parser.add_argument('hook', help='name of the hook to run')
        return parser

    def run(self, options: Namespace):
        hooks = get_hooks()
        hook = hooks.get(options.hook)
        if hook:
            hook(Snap())


script = HookScript()



from pathlib import Path

import pytest
import yaml

from ..snaphelpers import SnapHelpersScript


@pytest.fixture
def src_dir(tmpdir):
    src_dir = Path(tmpdir / 'src')
    src_dir.mkdir()
    yield src_dir


@pytest.fixture
def prime_dir(tmpdir):
    prime_dir = Path(tmpdir / 'prime')
    prime_dir.mkdir()
    yield prime_dir


@pytest.fixture
def snapcraft_env(monkeypatch, src_dir, prime_dir):
    monkeypatch.setenv('SNAPCRAFT_PART_SRC', str(src_dir))
    monkeypatch.setenv('SNAPCRAFT_PRIME', str(prime_dir))


@pytest.fixture
def snapcraft_yaml(src_dir):
    content = {'hooks': {'install': {}, 'configure': {}}}
    snap_dir = src_dir / 'snap'
    snap_dir.mkdir(parents=True)
    snapcraft_yaml = snap_dir / 'snapcraft.yaml'
    with snapcraft_yaml.open('w') as fd:
        yaml.dump(content, stream=fd)
    yield snapcraft_yaml


@pytest.mark.usefixtures('snapcraft_env')
class TestSnapHelpersScript:

    def test_write_hooks_missing_part_dir(self, monkeypatch):
        monkeypatch.delenv('SNAPCRAFT_PART_SRC')
        script = SnapHelpersScript()
        with pytest.raises(RuntimeError) as e:
            script(['write-hooks'])
        assert (
            'SNAPCRAFT_PART_SRC environment variable not defined' in str(
                e.value))

    def test_write_hooks_missing_prime_dir(self, monkeypatch):
        monkeypatch.delenv('SNAPCRAFT_PRIME')
        script = SnapHelpersScript()
        with pytest.raises(RuntimeError) as e:
            script(['write-hooks'])
        assert (
            'SNAPCRAFT_PRIME environment variable not defined' in str(e.value))

    def test_write_hooks_create_files(self, capsys, prime_dir, snapcraft_yaml):
        script = SnapHelpersScript()
        script(['write-hooks'])
        hooks_dir = prime_dir / 'snap' / 'hooks'
        configure_hook, install_hook = sorted(hooks_dir.iterdir())
        assert (
            '"${SNAP}/bin/snap-helpers-hook" "configure"' in
            configure_hook.read_text())
        assert (
            '"${SNAP}/bin/snap-helpers-hook" "install"' in
            install_hook.read_text())
        out = capsys.readouterr().out
        assert 'Writing hook files' in out
        assert f'configure -> {hooks_dir}/configure' in out
        assert f'install -> {hooks_dir}/install' in out

    def test_write_hooks_no_hooks(self, capsys, prime_dir, snapcraft_yaml):
        snapcraft_yaml.write_text('{}')
        script = SnapHelpersScript()
        script(['write-hooks'])
        hooks_dir = prime_dir / 'snap' / 'hooks'
        assert not hooks_dir.exists()
        assert 'No hooks defined in the snap' in capsys.readouterr().out



import json
from textwrap import dedent
from unittest.mock import call

import pytest

from .._ctl import (
    ServiceInfo,
    SnapCtl,
    SnapCtlError,
)


@pytest.mark.usefixtures('snap_apply_env')
class TestSnapCtl:

    def test_run(self, tmpdir):
        executable = tmpdir / 'snapctl'
        executable.write_text(
            dedent(
                f'''\
                #!/bin/sh
                echo foo bar
                '''), 'utf-8')
        executable.chmod(0o755)
        snapctl = SnapCtl(executable=str(executable))
        assert snapctl.run() == 'foo bar\n'

    def test_run_fail(self, tmpdir):
        executable = tmpdir / 'snapctl'
        executable.write_text(
            dedent(
                f'''\
                #!/bin/sh
                echo 'fail!' >&2
                exit 1
                '''), 'utf-8')
        executable.chmod(0o755)
        snapctl = SnapCtl(executable=str(executable))
        with pytest.raises(SnapCtlError) as e:
            snapctl.run()
        assert str(e.value) == 'Call to snapctl failed with error 1: fail!\n'

    def test_config_get(self, snapctl):
        output = {'foo': 123, 'bar': 'BAR'}
        snapctl.run.return_value = json.dumps(output)
        assert snapctl.config_get('foo', 'bar') == output
        assert snapctl.run.mock_calls == [call('get', '-d', 'foo', 'bar')]

    def test_config_set(self, snapctl):
        snapctl.config_set({'foo.bar': 123, 'baz': [1, 2, 3]})
        assert snapctl.run.mock_calls == [
            call('set', 'foo.bar=123', 'baz=[1, 2, 3]')
        ]

    def test_start(self, snapctl):
        snapctl.start()
        assert snapctl.run.mock_calls == [call('start', 'mysnap_inst')]

    def test_start_enable(self, snapctl):
        snapctl.start(enable=True)
        assert snapctl.run.mock_calls == [
            call('start', '--enable', 'mysnap_inst')
        ]

    def test_start_services(self, snapctl):
        snapctl.start('foo', 'bar')
        assert snapctl.run.mock_calls == [
            call('start', 'mysnap_inst.foo', 'mysnap_inst.bar')
        ]

    def test_stop(self, snapctl):
        snapctl.stop()
        assert snapctl.run.mock_calls == [call('stop', 'mysnap_inst')]

    def test_stop_enable(self, snapctl):
        snapctl.stop(disable=True)
        assert snapctl.run.mock_calls == [
            call('stop', '--disable', 'mysnap_inst')
        ]

    def test_stop_services(self, snapctl):
        snapctl.stop('foo', 'bar')
        assert snapctl.run.mock_calls == [
            call('stop', 'mysnap_inst.foo', 'mysnap_inst.bar')
        ]

    def test_restart(self, snapctl):
        snapctl.restart()
        assert snapctl.run.mock_calls == [call('restart', 'mysnap_inst')]

    def test_restart_enable(self, snapctl):
        snapctl.restart(reload=True)
        assert snapctl.run.mock_calls == [
            call('restart', '--reload', 'mysnap_inst')
        ]

    def test_restart_services(self, snapctl):
        snapctl.restart('foo', 'bar')
        assert snapctl.run.mock_calls == [
            call('restart', 'mysnap_inst.foo', 'mysnap_inst.bar')
        ]

    def test_services(self, snapctl):
        snapctl.run.return_value = dedent(
            f'''\
            Service          Startup   Current   Notes
            mysnap.service1  disabled  inactive  foo,bar
            mysnap.service2  enabled   active    -
            mysnap.service3  enabled   inactive  baz
            ''')
        assert snapctl.services() == [
            ServiceInfo(
                name='service1',
                enabled=False,
                active=False,
                notes=['foo', 'bar']),
            ServiceInfo(name='service2', enabled=True, active=True, notes=[]),
            ServiceInfo(
                name='service3', enabled=True, active=False, notes=['baz'])
        ]
        assert snapctl.run.mock_calls == [call('services', 'mysnap_inst')]

    def test_services_with_services(self, snapctl):
        snapctl.run.return_value = dedent(
            f'''\
            Service          Startup   Current   Notes
            mysnap.service1  disabled  inactive  foo,bar
            mysnap.service3  enabled   inactive  baz
            ''')
        assert snapctl.services('service1', 'service3') == [
            ServiceInfo(
                name='service1',
                enabled=False,
                active=False,
                notes=['foo', 'bar']),
            ServiceInfo(
                name='service3', enabled=True, active=False, notes=['baz'])
        ]
        assert snapctl.run.mock_calls == [
            call('services', 'mysnap_inst.service1', 'mysnap_inst.service3')
        ]



from argparse import (
    ArgumentParser,
    Namespace,
)

from .._script import Script


class SampleScript(Script):

    def get_parser(self):
        parser = ArgumentParser()
        parser.add_argument('foo')
        parser.add_argument('bar')
        return parser

    def run(self, options: Namespace):
        self.options = options


class TestScript:

    def test_run(self):
        script = SampleScript()
        script(args=['FOO', 'BAR'])
        assert script.options == Namespace(foo='FOO', bar='BAR')



from typing import (
    Callable,
    Dict,
)

import pkg_resources

from ._snap import Snap


def get_hooks(
        pkg_resources=pkg_resources) -> Dict[str, Callable[[Snap], None]]:
    """Return registered snap hooks.

    Resources registreed in setup.py as `snaphelpers.hooks` are loaded as
    hooks, based on their name.

    For example:

    .. code::

       setup(
           # ...
           entry_points={
               'snaphelpers.hooks': [
                   'install = 'foo.bar:install_hook',
                   'configure = 'foo.bar:configure_hook'
               ]
           }
       )

    """
    return {
        entry.name: entry.load()
        for entry in pkg_resources.iter_entry_points('snaphelpers.hooks')
    }



import collections
import os
from typing import (
    Iterator,
    Mapping,
    Optional,
)


def is_snap(environ: Optional[Mapping[str, str]] = None) -> bool:
    """Return whether running in a Snap."""
    if environ is None:
        environ = os.environ
    return bool(environ.get('SNAP', ''))


class SnapEnviron(collections.Mapping):
    """Environment variables related to the Snap.

    This provides read-only access to environment variables starting with
    :data:`SNAP_`.

    These can be accessed either as a dict or as attributes, without the
    :data:`SNAP_` prefix. E.g.::

      env = SnapEnviron()
      env.NAME     # -> 'mysnap'
      env['NAME']  # -> 'mysnap'

    *Note*: The :data:`SNAP` environment variable is also included.

    """

    _PREFIX = 'SNAP_'

    def __init__(self, environ: Optional[Mapping[str, str]] = None):
        if environ is None:
            environ = os.environ
        self._env = {
            key[len(self._PREFIX):]: value
            for key, value in environ.items() if key.startswith(self._PREFIX)
        }
        self._env['SNAP'] = environ['SNAP']

    def __getitem__(self, key: str) -> str:
        return self._env[key]

    def __len__(self) -> int:
        return len(self._env)

    def __iter__(self) -> Iterator:
        return iter(self._env)

    def __getattr__(self, attr: str) -> str:
        try:
            return self._env[attr]
        except KeyError as e:
            raise AttributeError(str(e))



from textwrap import dedent
from unittest.mock import call

import pytest

from .._ctl import (
    ServiceInfo,
    SnapCtl,
)
from .._service import (
    SnapService,
    SnapServices,
)


@pytest.fixture
def snap_service_status_output(snapctl):
    yield dedent(
        f'''\
        Service               Startup   Current   Notes
        mysnap_inst.serv1     disabled  inactive  foo,bar
        ''')


class TestSnapService:

    def test_attrs(self, snapctl):
        info = ServiceInfo(
            name='serv1', enabled=True, active=False, notes=['foo', 'bar'])
        service = SnapService(info, snapctl=snapctl)
        assert service.name == 'serv1'
        assert service.enabled
        assert not service.active
        assert service.notes == ['foo', 'bar']

    def test_eq(self, snapctl):
        info = ServiceInfo(
            name='serv1', enabled=True, active=False, notes=['foo', 'bar'])
        assert (
            SnapService(info, snapctl=snapctl) == SnapService(
                info, snapctl=snapctl))

    def test_eq_other_info(self, snapctl):
        info1 = ServiceInfo(
            name='serv1', enabled=True, active=False, notes=['foo', 'bar'])
        info2 = ServiceInfo(
            name='serv2', enabled=True, active=False, notes=['foo', 'bar'])
        assert (
            SnapService(info1, snapctl=snapctl) != SnapService(
                info2, snapctl=snapctl))

    def test_eq_other_snapctl(self, snapctl):
        info = ServiceInfo(
            name='serv1', enabled=True, active=False, notes=['foo', 'bar'])
        assert (
            SnapService(info, snapctl=snapctl) != SnapService(
                info, snapctl=SnapCtl()))

    @pytest.mark.parametrize('action', ['start', 'stop', 'restart'])
    def test_actions(self, action, snapctl, snap_service_status_output):
        snapctl.run.side_effect = ['', snap_service_status_output]
        info = ServiceInfo(name='serv1', enabled=True, active=True, notes=[])
        service = SnapService(info, snapctl=snapctl)
        getattr(service, action)()
        assert snapctl.run.mock_calls == [
            call(action, 'mysnap_inst.serv1'),
            call('services', 'mysnap_inst.serv1')
        ]

    @pytest.mark.parametrize(
        'action, option',
        [('start', 'enable'), ('stop', 'disable'), ('restart', 'reload')])
    def test_actions_with_options(
            self, action, option, snapctl, snap_service_status_output):
        snapctl.run.side_effect = ['', snap_service_status_output]
        info = ServiceInfo(name='serv1', enabled=True, active=True, notes=[])
        service = SnapService(info, snapctl=snapctl)
        getattr(service, action)(**{option: True})
        assert snapctl.run.mock_calls == [
            call(action, f'--{option}', 'mysnap_inst.serv1'),
            call('services', 'mysnap_inst.serv1')
        ]

    def test_refresh_status(self, snapctl, snap_service_status_output):
        snapctl.run.return_value = snap_service_status_output
        info = ServiceInfo(name='serv1', enabled=True, active=True, notes=[])
        service = SnapService(info, snapctl=snapctl)
        service.refresh_status()
        assert not service.enabled
        assert not service.active
        assert service.notes == ['foo', 'bar']


class TestSnapServices:

    @pytest.mark.parametrize('action', ['start', 'stop', 'restart'])
    def test_actions(self, action, snapctl):
        services = SnapServices(snapctl=snapctl)
        getattr(services, action)()
        assert snapctl.run.mock_calls == [call(action, 'mysnap_inst')]

    def test_list(self, fake_snapctl):
        services = SnapServices(snapctl=fake_snapctl)
        info1 = ServiceInfo(
            name='serv1', enabled=True, active=False, notes=['foo'])
        info2 = ServiceInfo(
            name='serv2', enabled=False, active=True, notes=['bar'])
        fake_snapctl._services = [info1, info2]
        assert services.list() == {
            'serv1': SnapService(info1, snapctl=fake_snapctl),
            'serv2': SnapService(info2, snapctl=fake_snapctl)
        }



import json
import re
from subprocess import (
    PIPE,
    Popen,
)
from typing import (
    Any,
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
)

from ._env import SnapEnviron


class ServiceInfo(NamedTuple):
    """Information for a service in the snap."""

    name: str
    enabled: bool
    active: bool
    notes: List[str]


class SnapCtlError(Exception):
    """A snapctl command failed."""

    returncode: int
    error: str

    def __init__(self, process: Popen):
        self.returncode = process.returncode
        self.error = process.stderr.read().decode('utf-8')
        super().__init__(
            f'Call to snapctl failed with error {self.returncode}: ' +
            self.error)


class SnapCtl:
    """Run the :data:`snapctl` command."""

    _SERVICE_RE = re.compile(
        r'[^.]+\.(?P<name>\S+)\s+'
        r'(?P<startup>\S+)\s+'
        r'(?P<current>\S+)\s+'
        r'(?P<notes>\S+)')

    def __init__(
            self,
            executable: str = '/usr/bin/snapctl',
            env: Optional[SnapEnviron] = None):
        if env is None:
            env = SnapEnviron()
        self._executable = executable
        self._instance_name = env.INSTANCE_NAME

    def start(self, *services: str, enable: bool = False):
        """Start all or specified services in the snap.

        :param services: a list of services defined in the snap to start.
          If not specified, all services will be started.
        :param enable: whether to also enable services at startup.

        """
        self._run_for_services('start', services, options={'enable': enable})

    def stop(self, *services: str, disable: bool = False):
        """Stop all or specified services in the snap.

        :param services: a list of services defined in the snap to stop.
          If not specified, all services will be stopped.
        :param disable: whether to also disable services at startup.

        """
        self._run_for_services('stop', services, options={'disable': disable})

    def restart(self, *services: str, reload: bool = False):
        """Restart all or specified services in the snap.

        :param services: a list of services defined in the snap to restart.
          If not specified, all services will be restarted.
        :param reload: whether to reload services if supported.

        """
        self._run_for_services('restart', services, options={'reload': reload})

    def services(self, *services: str) -> List[ServiceInfo]:
        """Return info about services in the snap.

        :param services: a list of services to return info for.
          If not specified, all services are returned.

        """
        lines = self._run_for_services('services', services)
        service_infos = []
        # skip header
        for line in lines.splitlines()[1:]:
            match = self._SERVICE_RE.match(line)
            if match:
                info = match.groupdict()
                notes: List[str] = []
                if info['notes'] != '-':
                    notes = info['notes'].split(',')
                service_infos.append(
                    ServiceInfo(
                        name=info['name'],
                        enabled=info['startup'] == 'enabled',
                        active=info['current'] == 'active',
                        notes=notes))
        return service_infos

    def config_get(self, *keys: str) -> Dict[str, Any]:
        """Return the snap config.

        :param keys: a list of config keys to return.

        """
        conf: Dict[str, Any]
        conf = json.loads(self.run('get', '-d', *keys))
        return conf

    def config_set(self, configs: Dict[str, Any]):
        """Set snap configs.

        :param configs: a dict with configs. Keys can use dotted notation.

        """
        args = [f'{key}={json.dumps(value)}' for key, value in configs.items()]
        self.run('set', *args)

    def run(self, *args: str) -> str:
        """Execute the command return its output.

        ":param args: command args.

        """
        process = Popen([self._executable, *args], stdout=PIPE, stderr=PIPE)
        process.wait()
        if process.returncode:
            raise SnapCtlError(process)
        output: bytes = process.stdout.read()
        return output.decode('utf-8')

    def _run_for_services(
            self,
            cmd: str,
            services: Sequence[str],
            options: Optional[Dict[str, bool]] = None) -> str:
        opts: List[str] = []
        if options:
            opts = [
                f'--{option}' for option, value in options.items() if value
            ]
        if services:
            service_names = [
                f'{self._instance_name}.{service}' for service in services
            ]
        else:
            service_names = [self._instance_name]
        return self.run(cmd, *opts, *service_names)



from typing import (
    Mapping,
    Optional,
)

from ._conf import SnapConfig
from ._ctl import SnapCtl
from ._env import SnapEnviron
from ._path import SnapPaths
from ._service import SnapServices


class EnvironProperty:
    """Wrapper to get properties from a :class:`SanpEnviron` instance."""

    def __init__(self, name: str):
        self.name = name

    def __get__(self, instance, owner):
        return getattr(instance.environ, self.name)


class Snap:
    """Top-level wrapper for a Snap."""

    config: SnapConfig
    environ: SnapEnviron
    paths: SnapPaths
    services: SnapServices

    def __init__(self, environ: Optional[Mapping[str, str]] = None):
        self.environ = SnapEnviron(environ=environ)
        self.paths = SnapPaths(env=self.environ)
        snapctl = SnapCtl(env=self.environ)
        self.config = SnapConfig(snapctl=snapctl)
        self.services = SnapServices(snapctl=snapctl)

    def __str__(self):
        return (
            f'{self.__class__.__name__}'
            f'({self.name} {self.version} {self.revision})')

    name = EnvironProperty('NAME')
    instance_name = EnvironProperty('INSTANCE_NAME')
    version = EnvironProperty('VERSION')
    revision = EnvironProperty('REVISION')



from pathlib import Path

from .._path import SnapPaths


class TestSnapPaths:

    def test_default_env(self, snap_apply_env):
        paths = SnapPaths()
        assert paths.common == Path('/var/snap/mysnap/common')
        assert paths.data == Path('/var/snap/mysnap/123')
        assert paths.snap == Path('/snap/mysnap/123')
        assert paths.user_common == Path('/home/user/snap/mysnap/common')
        assert paths.user_data == Path('/home/s/snap/mysnap/123')

    def test_paths(self, snap_environ):
        paths = SnapPaths(env=snap_environ)
        assert paths.common == Path('/var/snap/mysnap/common')
        assert paths.data == Path('/var/snap/mysnap/123')
        assert paths.snap == Path('/snap/mysnap/123')
        assert paths.user_common == Path('/home/user/snap/mysnap/common')
        assert paths.user_data == Path('/home/s/snap/mysnap/123')

    def test_repr(self, snap_environ):
        paths = SnapPaths(env=snap_environ)
        assert repr(paths) == (
            'SnapPaths('
            'common=/var/snap/mysnap/common '
            'data=/var/snap/mysnap/123 '
            'snap=/snap/mysnap/123 '
            'user_common=/home/user/snap/mysnap/common '
            'user_data=/home/s/snap/mysnap/123)')



import pytest

from .. import hook
from ..hook import HookScript


@pytest.fixture
def hook_calls():
    yield []


@pytest.fixture
def hooks(hook_calls):
    yield {
        'configure': lambda snap: hook_calls.append('configure'),
        'install': lambda snap: hook_calls.append('install')
    }


@pytest.fixture
def mock_get_hooks(mocker, hooks):
    mocker.patch.object(hook, 'get_hooks', lambda: hooks)


@pytest.mark.usefixtures('snap_apply_env', 'mock_get_hooks')
class TestHookScript:

    def test_run_hook(self, hook_calls):
        script = HookScript()
        script(args=['install'])
        script(args=['configure'])
        assert hook_calls == ['install', 'configure']

    def test_no_hook(self, hooks, hook_calls):
        hooks.pop('install')
        script = HookScript()
        script(args=['install'])
        assert hook_calls == []



from pathlib import Path
from typing import Optional

from ._env import SnapEnviron


class SnapPaths:
    """Paths related to the snap."""

    __slots__ = ('common', 'data', 'snap', 'user_common', 'user_data')

    common: Path  #: the SNAP_COMMON path
    data: Path  #: the SNAP_DATA path
    snap: Path  #: the SNAP path
    user_common: Path  #: the SNAP_USER_COMMON path
    user_data: Path  #: the SNAP_USER_DATA path

    def __init__(self, env: Optional[SnapEnviron] = None):
        if env is None:
            env = SnapEnviron()
        for key in self.__slots__:
            setattr(self, key, Path(env[key.upper()]))

    def __repr__(self):
        name = self.__class__.__name__
        values = ' '.join(
            f'{key}={str(getattr(self, key))}' for key in self.__slots__)
        return f'{name}({values})'



from abc import (
    ABC,
    abstractmethod,
)
from argparse import (
    ArgumentParser,
    Namespace,
)
from typing import (
    List,
    Optional,
)


class Script(ABC):
    """A Script."""

    @abstractmethod
    def get_parser(self) -> ArgumentParser:
        """Return a parser for the script.

        Subclasses must implement this method.
        """

    @abstractmethod
    def run(self, options: Namespace):
        """Run the script

        Subclasses must implement this method.

        """

    def __call__(self, args: Optional[List[str]] = None):
        parser = self.get_parser()
        opts = parser.parse_args(args=args)
        self.run(opts)



from pathlib import Path

import pytest

from .._snap import Snap


class TestSnap:

    def test_str(self, snap_env):
        snap = Snap(environ=snap_env)
        assert str(snap) == 'Snap(mysnap 0.1.2 123)'

    @pytest.mark.parametrize(
        'name', ['name', 'instance_name', 'version', 'revision'])
    def test_properties(self, name, snap_env):
        snap = Snap(environ=snap_env)
        assert getattr(snap, name) == snap_env[f'SNAP_{name.upper()}']

    def test_paths(self, snap_env):
        snap = Snap(environ=snap_env)
        assert snap.paths.snap == Path('/snap/mysnap/123')
        assert snap.paths.common == Path('/var/snap/mysnap/common')



import pytest

from .._conf import (
    InvalidKey,
    SnapConfig,
    SnapConfigOptions,
    UnknownConfigKey,
)


class TestSnapConfigOptions:

    def test_fetch(self, snap_config, fake_snapctl):
        options = SnapConfigOptions(['foo', 'baz'], snapctl=fake_snapctl)
        options.fetch()
        assert options._config == {
            'foo': 123,
            'baz': {
                'aaa': 'nested',
                'bbb': {
                    'ccc': 'more nested'
                }
            }
        }

    def test_as_dict(self, snap_config, fake_snapctl):
        options = SnapConfigOptions(['foo', 'baz'], snapctl=fake_snapctl)
        options.fetch()
        options.as_dict() == snap_config

    def test_as_dict_not_fetched(self, fake_snapctl):
        options = SnapConfigOptions(['foo', 'baz'], snapctl=fake_snapctl)
        options.as_dict() == {}

    @pytest.mark.parametrize(
        'key,value', [
            ('foo', 123), ('baz.aaa', 'nested'),
            ('baz.bbb.ccc', 'more nested'), ('blah', [1, 2, 3])
        ])
    def test_getitem(self, key, value, fake_snapctl):
        options = SnapConfigOptions(
            ['foo', 'baz', 'blah'], snapctl=fake_snapctl)
        options.fetch()
        assert options[key] == value

    @pytest.mark.parametrize('key', ['unknown', 'baz.unknown'])
    def test_getitem_notfound(self, key, fake_snapctl):
        options = SnapConfigOptions(['foo', 'baz'], snapctl=fake_snapctl)
        options.fetch()
        with pytest.raises(UnknownConfigKey) as e:
            options[key]
        assert e.value.key == key

    def test_getitem_not_fetched(self, fake_snapctl):
        options = SnapConfigOptions(['foo', 'baz'], snapctl=fake_snapctl)
        with pytest.raises(UnknownConfigKey):
            options['foo']

    @pytest.mark.parametrize(
        'key,contained', [
            ('foo', True), ('nope', False), ('baz.bbb', True),
            ('baz.nope', False), ('baz.bbb.ccc', True),
            ('baz.bbb.nope', False)
        ])
    def test_in(self, key, contained, fake_snapctl):
        options = SnapConfigOptions(['foo', 'baz'], snapctl=fake_snapctl)
        options.fetch()
        assert (key in options) == contained

    @pytest.mark.parametrize(
        'key,value', [
            ('foo', 123), ('baz.aaa', 'nested'),
            ('baz.bbb.ccc', 'more nested'), ('blah', [1, 2, 3])
        ])
    def test_get(self, key, value, fake_snapctl):
        options = SnapConfigOptions(
            ['foo', 'baz', 'blah'], snapctl=fake_snapctl)
        options.fetch()
        assert options.get(key, None) == value

    def test_get_default(self, fake_snapctl):
        options = SnapConfigOptions(
            ['foo', 'baz', 'blah'], snapctl=fake_snapctl)
        options.fetch()
        assert options.get('something.else', 'this') == 'this'


class TestSnapConfig:

    @pytest.mark.parametrize(
        'key,value', [
            ('foo', 123), ('baz.aaa', 'nested'),
            ('baz.bbb.ccc', 'more nested'), ('blah', [1, 2, 3])
        ])
    def test_config_get(self, key, value, fake_snapctl):
        config = SnapConfig(snapctl=fake_snapctl)
        assert config.get(key) == value

    @pytest.mark.parametrize(
        'key', ['not-here', 'foo.not-here', 'baz.aaa.not-here'])
    def test_config_get_not_found(self, key, fake_snapctl):
        config = SnapConfig(snapctl=fake_snapctl)
        with pytest.raises(UnknownConfigKey):
            config.get(key)

    @pytest.mark.parametrize(
        'key,value', [
            ('foo', 123), ('baz.aaa', 'nested'),
            ('baz.bbb.ccc', 'more nested'), ('blah', [1, 2, 3])
        ])
    def test_get_options(self, key, value, fake_snapctl):
        config = SnapConfig(snapctl=fake_snapctl)
        options = config.get_options('foo', 'baz', 'blah')
        assert options[key] == value

    def test_get_options_only_top_level(self, fake_snapctl):
        config = SnapConfig(snapctl=fake_snapctl)
        with pytest.raises(InvalidKey) as e:
            config.get_options('foo', 'baz.bar')
        assert e.value.key == 'baz.bar'

    def test_set(self, fake_snapctl):
        config = SnapConfig(snapctl=fake_snapctl)
        config.set({'one': 1, 'two': {'three': 3}})
        assert fake_snapctl._configs['one'] == 1
        assert fake_snapctl._configs['two'] == {'three': 3}
        options = config.get_options('one', 'two')
        assert options.as_dict() == {'one': 1, 'two': {'three': 3}}



from copy import deepcopy
from typing import (
    Any,
    Dict,
    Optional,
    Sequence,
)

from ._ctl import SnapCtl


class UnknownConfigKey(Exception):
    """The specified config key is unknown."""

    key: str

    def __init__(self, key: str):
        self.key = key
        super().__init__(f'Unknown config key: {key}')


class InvalidKey(Exception):
    """The specified key is invalid."""

    key: str

    def __init__(self, key: str):
        self.key = key
        super().__init__(f'Invalid top-level key: {key}')


class SnapConfigOptions:
    """Allow accessing a set of Snap config options with a dict-like interface.

    Nested keys can be accessed using dotted notation::

        config['foo.bar.baz']

    :param keys: the top-level configuration keys.

    """

    _config: Optional[Dict[str, Any]] = None

    def __init__(self, keys: Sequence[str], snapctl: Optional[SnapCtl] = None):
        self._keys = list(keys)
        self._snapctl = snapctl or SnapCtl()

    def __getitem__(self, item: str) -> Any:
        """Return value for a configuration key."""
        config = self._config
        for key in item.split('.'):
            if not isinstance(config, dict):
                raise UnknownConfigKey(item)
            try:
                config = config[key]
            except KeyError:
                raise UnknownConfigKey(item)
        return config

    def __contains__(self, item: str) -> bool:
        """Whether the configuration conains a key."""
        try:
            self[item]
        except UnknownConfigKey:
            return False
        return True

    def get(self, key: str, default: Any = None) -> Any:
        """Return value for a key, with a default."""
        try:
            return self[key]
        except UnknownConfigKey:
            return default

    def fetch(self):
        """Fetch (or refresh) configuration for the set of keys."""
        self._config = self._snapctl.config_get(*self._keys)

    def as_dict(self) -> Dict[str, Any]:
        """Return the configuration as a :class:`dict`."""
        if self._config is None:
            return {}
        return deepcopy(self._config)


class SnapConfig:
    """Interact with the snap configuration.

    It allows getting and setting configuration options.

    """

    def __init__(self, snapctl: Optional[SnapCtl] = None):
        self._snapctl = snapctl or SnapCtl()

    def get_options(self, *keys: str) -> SnapConfigOptions:
        """Return a :data:`SnapConfigOptions` for the specified keys."""
        for key in keys:
            if '.' in key:
                raise InvalidKey(key)
        options = SnapConfigOptions(keys=keys, snapctl=self._snapctl)
        options.fetch()
        return options

    def get(self, key: str) -> Any:
        """Return value for a single key.

        It raises :class:`UnknownConfigKey` if the option doesn't exist.

        """
        top_key = key.split('.', maxsplit=1)[0]
        options = self.get_options(top_key)
        return options[key]

    def set(self, options: Dict[str, Any]):
        """Set config options.

        :param options: a dict with configs. Keys can use dotted notation.

        """
        self._snapctl.config_set(options)



from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    Namespace,
)
import os
from pathlib import Path
from typing import (
    Mapping,
    Optional,
)

import yaml

from ._script import Script

HOOK_TEMPLATE = '''#!/bin/sh

exec "${{SNAP}}/snap/command-chain/snapcraft-runner" \
"${{SNAP}}/bin/snap-helpers-hook" "{hookname}"
'''


class SnapHelpersScript(Script):
    """Tool to perform snap-helpers tasks."""

    def __init__(self, environ: Optional[Mapping[str, str]] = None):
        if environ is None:
            environ = os.environ

    def get_parser(self) -> ArgumentParser:
        parser = ArgumentParser(
            description='Tool to perform snap-helpers tasks')
        subparsers = parser.add_subparsers(
            metavar='ACTION', dest='action', help='action to perform')
        subparsers.required = True

        subparsers.add_parser(
            'write-hooks',
            help='Write hook files',
            formatter_class=ArgumentDefaultsHelpFormatter)
        return parser

    def run(self, options: Namespace):
        action = options.action.replace('-', '_')
        getattr(self, f'_action_{action}')(options)

    def _action_write_hooks(self, options: Namespace):
        src_dir = self._ensure_env_path('SNAPCRAFT_PART_SRC')
        prime_dir = self._ensure_env_path('SNAPCRAFT_PRIME')

        with (src_dir / 'snap' / 'snapcraft.yaml').open() as fd:
            content = yaml.load(fd)

        hooks = content.get('hooks')
        if not hooks:
            print('No hooks defined in the snap.')
            return

        hooks_dir = prime_dir / 'snap' / 'hooks'
        if not hooks_dir.exists():
            hooks_dir.mkdir(parents=True)

        print('Writing hook files...')
        for hookname in hooks:
            hook_file = hooks_dir / hookname
            print(f' {hookname} -> {hook_file.absolute()}')
            hook_file.write_text(HOOK_TEMPLATE.format(hookname=hookname))
            hook_file.chmod(0o755)

    def _ensure_env_path(self, name: str) -> Path:
        value = os.environ.get(name)
        if value is None:
            raise RuntimeError(f'{name} environment variable not defined')
        return Path(value)


script = SnapHelpersScript()



from typing import (
    Callable,
    NamedTuple,
)

import pytest

from .._hook import get_hooks


class MockEntryPoint(NamedTuple):

    name: str
    load: Callable


@pytest.fixture
def hook_calls():
    yield []


@pytest.fixture
def entry_points(hook_calls):

    def make_load(name):

        def load():
            hook_calls.append(name)
            return f'loaded-{name}'

        return load

    yield [
        MockEntryPoint(name='install', load=make_load('install')),
        MockEntryPoint(name='configure', load=make_load('configure'))
    ]


@pytest.fixture
def mock_pkg_resources(mocker, entry_points):
    pkg_resources = mocker.Mock()
    pkg_resources.iter_entry_points.return_value = iter(entry_points)
    yield pkg_resources


class TestGetHooks:

    def test_hooks(self, mock_pkg_resources, hook_calls):
        assert get_hooks(pkg_resources=mock_pkg_resources) == {
            'configure': 'loaded-configure',
            'install': 'loaded-install'
        }
        assert hook_calls == ['install', 'configure']



import pytest

from .._env import (
    is_snap,
    SnapEnviron,
)


class TestIsSnap:

    @pytest.mark.parametrize(
        'value,expected', [('/snap/mysnap/12', True), ('', False)])
    def test_with_env(self, value, expected):
        environ = {'SNAP': value}
        assert is_snap(environ=environ) == expected

    def test_default_env(self, snap_apply_env):
        assert is_snap()


class TestSnapEnviron:

    def test_to_dict(self, snap_env):
        # ignored environment variables
        snap_env.update({'OTHER': 'value', 'PATH': '/bin:/usr/bin'})
        env = SnapEnviron(environ=snap_env)
        assert dict(env) == {
            'NAME': 'mysnap',
            'COMMON': '/var/snap/mysnap/common',
            'DATA': '/var/snap/mysnap/123',
            'INSTANCE_NAME': 'mysnap_inst',
            'REVISION': '123',
            'SNAP': '/snap/mysnap/123',
            'USER_COMMON': '/home/user/snap/mysnap/common',
            'USER_DATA': '/home/s/snap/mysnap/123',
            'VERSION': '0.1.2'
        }

    def test_len(self, snap_env):
        env = SnapEnviron(environ=snap_env)
        assert len(env) == len(snap_env)

    def test_getitem(self, snap_env):
        env = SnapEnviron(environ=snap_env)
        assert env['DATA'] == '/var/snap/mysnap/123'

    def test_setitem_not_allowed(self, snap_env):
        env = SnapEnviron(environ=snap_env)
        with pytest.raises(TypeError):
            env['FOO'] = 'bar'

    def test_getattr(self, snap_env):
        env = SnapEnviron(environ=snap_env)
        assert env.VERSION == '0.1.2'

    def test_getattr_not_found(self, snap_env):
        env = SnapEnviron(environ=snap_env)
        with pytest.raises(AttributeError):
            env.FOO

    def test_default_env(self, snap_env, snap_apply_env):
        assert SnapEnviron().SNAP == snap_env['SNAP']



from pathlib import Path

from setuptools import (
    find_packages,
    setup,
)

tests_require = ['pytest', 'pytest-mock']

config = {
    'name': 'snap-helpers',
    'version': '0.1.2',
    'license': 'LGPLv3+',
    'description': (
        'Helpers for interacting with the Snap system within a Snap'),
    'long_description': Path('README.rst').read_text(),
    'author': 'Alberto Donato',
    'author_email': 'alberto.donato@gmail.com',
    'maintainer': 'Alberto Donato',
    'maintainer_email': 'alberto.donato@gmail.com',
    'url': 'https://github.com/albertodonato/snap-helpers',
    'download_url': 'https://pypi.org/project/snap-helpers/#files',
    'packages': find_packages(include=['snaphelpers', 'snaphelpers.*']),
    'include_package_data': True,
    'entry_points': {
        'console_scripts': [
            'snap-helpers = snaphelpers.scripts.snaphelpers:script',
            'snap-helpers-hook = snaphelpers.scripts.hook:script'
        ]
    },
    'test_suite': 'snaphelpers',
    'install_requires': ['packaging', 'PyYAML'],
    'tests_require': tests_require,
    'extras_require': {
        'testing': tests_require
    },
    'keywords': 'snap snappy snapcraft',
    'classifiers': [
        'Development Status :: 4 - Beta',
        (
            'License :: OSI Approved :: '
            'GNU Lesser General Public License v3 or later (LGPLv3+)'),
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: System :: Archiving :: Packaging'
    ]
}

setup(**config)



"""Helpers for interacting with the Snap system within a Snap."""

from pkg_resources import get_distribution

from packaging.version import parse

from ._conf import (
    SnapConfig,
    SnapConfigOptions,
)
from ._ctl import (
    SnapCtl,
    SnapCtlError,
)
from ._env import (
    is_snap,
    SnapEnviron,
)
from ._path import SnapPaths
from ._service import SnapServices
from ._snap import Snap

__all__ = [
    'is_snap',
    'Snap',
    'SnapCtl',
    'SnapCtlError',
    'SnapConfig',
    'SnapConfigOptions',
    'SnapEnviron',
    'SnapPaths',
    'SnapServices',
    '__version__',
]

__version__ = parse(get_distribution('snap-helpers').version)



from typing import (
    Dict,
    Optional,
)

from ._ctl import (
    ServiceInfo,
    SnapCtl,
)


class SnapService:
    """A service defined in the Snap."""

    def __init__(self, info: ServiceInfo, snapctl: Optional[SnapCtl] = None):
        self._info = info
        self._snapctl = snapctl or SnapCtl()

    def __getattr__(self, attr):
        # forward attributes defined in ServiceInfo
        return getattr(self._info, attr)

    def __eq__(self, other) -> bool:
        return self._info == other._info and self._snapctl is other._snapctl

    def start(self, enable: bool = False):
        """Start the service.

        :param enable: whether to also enable the service at startup.

        """
        self._snapctl.start(self.name, enable=enable)
        self.refresh_status()

    def stop(self, disable: bool = False):
        """Stop the service.

        :param disable: whether to also disable the service at startup.

        """
        self._snapctl.stop(self.name, disable=disable)
        self.refresh_status()

    def restart(self, reload: bool = False):
        """Restart the service.

        :param reload: whether to reload the service if supported.

        """
        self._snapctl.restart(self.name, reload=reload)
        self.refresh_status()

    def refresh_status(self):
        """Update the status of the ervice service."""
        [self._info] = self._snapctl.services(self.name)


class SnapServices:
    """Manage services in the snap."""

    def __init__(self, snapctl: Optional[SnapCtl] = None):
        self._snapctl = snapctl or SnapCtl()

    def list(self) -> Dict[str, SnapService]:
        """Return services by name."""
        return {
            info.name: SnapService(info, snapctl=self._snapctl)
            for info in self._snapctl.services()
        }

    def start(self, enable: bool = False):
        """Start all services.

        :param enable: whether to also enable services at startup.

        """
        self._snapctl.start(enable=enable)

    def stop(self, disable: bool = False):
        """Stop all services.

        :param disable: whether to also disable services at startup.

        """
        self._snapctl.stop(disable=disable)

    def restart(self, reload: bool = False):
        """Restart all services.

        :param reload: whether to reload services if supported.

        """
        self._snapctl.restart(reload=reload)
